"""
TODO:
 - OK.. this is useful for non-system-daemons too... which have no pidfile..
 - split out pidfile concern to.. identity?
 - async[io] support, really just waiting
 - helpers for http, json, etc
 - heartbeat? status checks? heartbeat file?
 - zmq
 - cli, cfg reload
 - bootstrap
  - rsrc limit
  - logs
 - https://github.com/Homebrew/homebrew-services
 - deathpacts
 - timebomb
 - pickle protocol, revision / venv check, multiprocessing manager support
"""
import logging
import os.path
import time

from .. import check
from .. import dataclasses as dc
from .. import lang
from ..os.pidfiles.pidfile import Pidfile
from .launching import Launcher
from .spawning import Spawning
from .targets import Target
from .waiting import Wait
from .waiting import waiter_for


log = logging.getLogger(__name__)


##


class Daemon:
    @dc.dataclass(frozen=True, kw_only=True)
    class Config:
        target: Target
        spawning: Spawning

        #

        pid_file: str | None = None

        #

        reparent_process: bool = False
        launched_timeout_s: float = 5.

        #

        wait: Wait | None = None

        wait_timeout: lang.TimeoutLike = 10.
        wait_sleep_s: float = .1

        #

        def __post_init__(self) -> None:
            check.isinstance(self.pid_file, (str, None))

    def __init__(self, config: Config) -> None:
        super().__init__()

        self._config = config

    @property
    def config(self) -> Config:
        return self._config

    #

    @property
    def has_pidfile(self) -> bool:
        return self._config.pid_file is not None

    def _non_inheritable_pidfile(self) -> Pidfile:
        check.state(self.has_pidfile)
        return Pidfile(
            check.non_empty_str(self._config.pid_file),
            inheritable=False,
        )

    def is_running(self) -> bool:
        check.state(self.has_pidfile)

        if not os.path.isfile(check.non_empty_str(self._config.pid_file)):
            return False

        with self._non_inheritable_pidfile() as pf:
            return not pf.try_acquire_lock()

    #

    def wait_sync(self, timeout: lang.TimeoutLike = lang.Timeout.Default) -> None:
        if self._config.wait is None:
            return

        timeout = lang.Timeout.of(timeout, self._config.wait_timeout)
        waiter = waiter_for(self._config.wait)
        while not waiter.do_wait():
            timeout()
            time.sleep(self._config.wait_sleep_s or 0.)

    #

    def launch_no_wait(self) -> None:
        launcher = Launcher(
            target=self._config.target,
            spawning=self._config.spawning,

            pid_file=self._config.pid_file,
            reparent_process=self._config.reparent_process,
            launched_timeout_s=self._config.launched_timeout_s,
        )

        launcher.launch()

    def launch(self, timeout: lang.TimeoutLike = lang.Timeout.Default) -> None:
        self.launch_no_wait()

        self.wait_sync(timeout)



import abc
import functools
import os
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang


if ta.TYPE_CHECKING:
    import runpy
else:
    runpy = lang.proxy_import('runpy')


##


class Target(dc.Case):
    pass


class TargetRunner(abc.ABC):
    @abc.abstractmethod
    def run(self) -> None:
        raise NotImplementedError


@functools.singledispatch
def target_runner_for(target: Target) -> TargetRunner:
    raise TypeError(target)


##


class FnTarget(Target):
    fn: ta.Callable[[], None]


class FnTargetRunner(TargetRunner, dc.Frozen):
    target: FnTarget

    def run(self) -> None:
        self.target.fn()


@target_runner_for.register
def _(target: FnTarget) -> FnTargetRunner:
    return FnTargetRunner(target)


##


class NameTarget(Target):
    name: str


class NameTargetRunner(TargetRunner, dc.Frozen):
    target: NameTarget

    def run(self) -> None:
        name = self.target.name
        if lang.can_import(name):
            runpy._run_module_as_main(name)  # type: ignore  # noqa
        else:
            obj = lang.import_attr(self.target.name)
            obj()


##


class ExecTarget(Target):
    cmd: ta.Sequence[str] = dc.xfield(coerce=check.of_not_isinstance(str))


class ExecTargetRunner(TargetRunner, dc.Frozen):
    target: ExecTarget

    def run(self) -> None:
        os.execl(*self.target.cmd)


@target_runner_for.register
def _(target: ExecTarget) -> ExecTargetRunner:
    return ExecTargetRunner(target)



import os
import sys


def reparent_process() -> None:
    if (pid := os.fork()):  # noqa
        sys.exit(0)
        raise RuntimeError('Unreachable')  # noqa

    os.setsid()

    if (pid := os.fork()):  # noqa
        sys.exit(0)

    sys.stdout.flush()
    sys.stderr.flush()
